@@ -5,14 +5,16 @@
 # License: MIT
 ##
 
-from .model import Model, ExtendModule
+from .model import Model, ExtendModel, PreparedModel
 from .train import TrainStatus, TrainVerboseCallback, do_train
 
-__version__ = "0.0.9"
+__version__ = "0.0.11"
 
 __all__ = (
 	
 	"Model",
+	"ExtendModel",
+	"PreparedModel",
 	"TrainStatus",
 	"TrainVerboseCallback",
 	"do_train",
